"""
This module contains the terminal user interface and the webserver.

### Submodules

- **terminal**: Terminal user interface
- **external**: Functions for checking and downloading external dependencies (binaries and data)
- **config**: Configuration file and command line arguments
- **web**: Webserver
"""